     /**
      * \brief JSON : Checker Utilities
      *
      * Utility module for validating string for the JSON syntax.
      * The strings are converted from the passed or Job ccsid to
      * ASCII (ccsid 819). Comments before, inline and after the JSON
      * string are not supported.
      *
      * \author Mihael Schmidt
      * \date   2009-02-28
      *
      * \link http://www.json.org/JSON_checker/ JSON Checker
      * 
      * \rev 2012-08-18 Brian J. Garland
      *      fixed bugs in json_validate and json_validateFile
      * 
      * \rev 2014-03-27 Paul Foster
      *      fixed bug in json_validateFile - checking large files 
      */

      *-------------------------------------------------------------------------
      * And thanks to Scott Klement for his excellent tutorials about IFS and
      * iconv() at http://www.scottklement.com .
      *-------------------------------------------------------------------------

      *-------------------------------------------------------------------------
      * Copyright (c) 2007-2014 Mihael Schmidt
      * All rights reserved.
      *
      * This file is part of the JSON service program.
      *
      * JSON service program is free software: you can redistribute it and/or modify it under
      * the terms of the GNU Lesser General Public License as published by
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * JSON service program is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public
      * License along with JSON service program.  If not, see http://www.gnu.org/licenses/.
      *-------------------------------------------------------------------------

     H nomain
     H bnddir('QC2LE')
     H copyright('Copyright (c) 2007-2014 Mihael Schmidt. All rights reserved.')


      *------------------------------------------------------------------------
      * Prototypes
      *------------------------------------------------------------------------
      /copy 'json_h.rpgle'
      /copy 'json_int_h.rpgle'
      /copy 'iconv_h.rpgle'
      /copy 'ifsio_h.rpgle'
      /copy 'libc_h.rpgle'
      /copy 'ceeapi_h.rpgle'


      *-------------------------------------------------------------------------
      * Global Variables
      *-------------------------------------------------------------------------
      *
      * Parameters passed to CEEDOD
      *
     D cee_descType    S             10I 0
     D cee_dataType    S             10I 0
     D cee_descInfo1   S             10I 0
     D cee_descInfo2   S             10I 0
     D cee_length      S             10I 0
      *
     D true            S             10I 0 inz(1)
     D false           S             10I 0 inz(0)

      *------------------------------------------------------------------------
      * Procedures
      *------------------------------------------------------------------------

     /**
      * \brief Validate JSON string
      *
      * Validates the passed JSON string.
      *
      * \author Mihael Schmidt
      *
      * \param JSON string
      * \param CCSID from the source string (default: 0 = Job CCSID)
      *
      * \return *on = valid JSON string <br>
      *         *off = invalid JSON string
      */
     P json_validate   B                   export
     D                 PI              N   opdesc
     D   string                   65535A   const options(*varsize)
     D   ccsid                       10I 0 const options(*nopass)
      *
     D iconv_table     DS                  likeds(iconv_t)
     D json_string_ptr...
     D                 S               *
     D json_string     S          65535A
     D json_string_length...
     D                 S             10U 0
     D json_checker_ptr...
     D                 S               *
     D json_checker    DS                  likeds(tmpl_json_checker)
     D                                     based(json_checker_ptr)
      *
     D conv_ds         DS
     D unsInt                        10U 0 inz(0)
     D char                           1A   overlay(unsInt : 4)
      *
     D tmpLength       S             10U 0
     D i               S             10I 0
     D valid           S               N   inz(*on)
      /free
       cee_getOpDescInfo(1 : cee_descType : cee_dataType : cee_descInfo1 :
                         cee_descInfo2 : cee_length : *omit);
       json_string = %trim(%subst(string : 1 : cee_length)) + x'00';
       json_string_ptr = %addr(json_string);
       json_string_length = strlen(json_string_ptr);
       tmpLength = json_string_length;

       if (%parms() = 2);
         iconv_from.ccsid = ccsid;
       else;
         iconv_from.ccsid = 0;
       endif;

       //
       // translate the string from the current job ccsid (or passed ccsid)
       // to ascii (819) because the checker only understands ascii
       //
       iconv_table = iconv_open(iconv_to : iconv_from);
       if (iconv_table.return_value <> -1);

         if (iconv(iconv_table :
                   json_string_ptr : tmpLength :
                   json_string_ptr : tmpLength) <> -1);

           // create json checker
           json_checker_ptr = json_new_checker(99);

           // check string char by char
           for i = 1 to json_string_length;
             char =  %subst(json_string : i : 1);
             if (json_checker_char(json_checker_ptr : unsInt) = false);
               sendDiagnosticMessage('JSON checker aborted at position ' +
                     %char(i) + '.');
               valid = *off;
               leave;
             endif;
           endfor;

           // free memory
           if (valid and json_checker_done(json_checker_ptr) = false);
             valid = *off;
           endif;

         else;
           sendDiagnosticMessage('Could not translate buffer from ccsid ' +
                %char(iconv_from.ccsid) + ' to ccsid ' +
                %char(iconv_to.ccsid) + '.');
           valid = *off;
         endif;

         // cleanup
         iconv_close(iconv_table);

       else;
         sendDiagnosticMessage('Could not init iconv.');
         valid = *off;
       endif;

       return valid;
      /end-free
     P                 E


     /**
      * \brief Validate JSON file
      *
      * Validates the JSON string in the passed file. The file should only
      * contain the JSON string. Comments are not supported.
      *
      * \author Mihael Schmidt
      *
      * \param IFS file path
      * \param CCSID from the source string (default: 0 = Job CCSID)
      *
      * \return *on = valid JSON string <br>
      *         *off = invalid JSON string
      */
     P json_validateFile...
     P                 B                   export
     D                 PI              N   opdesc
     D   path                     65535A   const options(*varsize)
     D   ccsid                       10I 0 const options(*nopass)
      *
      * file stuff
      *
     D filePath        S          65535A
     D fd              S             10I 0
     D dataPtr         S               *
     D data            S           1024A   inz
     D length          S             10I 0
      *
      * translation stuff
      *
     D iconv_table     DS                  likeds(iconv_t)
     D json_string_ptr...
     D                 S               *
     D json_string     S          65535A
     D json_checker_ptr...
     D                 S               *
     D json_checker    DS                  likeds(tmpl_json_checker)
     D                                     based(json_checker_ptr)
     D json_pos        S             10I 0 inz    
      *
     D conv_ds         DS
     D unsInt                        10U 0 inz(0)
     D char                           1A   overlay(unsInt : 4)
      *
     D tmpLength       S             10U 0
     D i               S             10I 0
     D valid           S               N   inz(*on)
      /free
       cee_getOpDescInfo(1 : cee_descType : cee_dataType : cee_descInfo1 :
                         cee_descInfo2 : cee_length : *omit);
       filePath = %trim(%subst(path : 1 : cee_length)) + x'00';
       json_string_ptr = %addr(json_string);
       dataPtr = %addr(data);

       fd = open(filePath : O_RDONLY + O_TEXTDATA);
       if (fd >= 0);

         if (%parms() = 2);
           iconv_from.ccsid = ccsid;
         else;
           iconv_from.ccsid = 0;
         endif;

         // create json checker
         json_checker_ptr = json_new_checker(99);

         //
         // translate the string from the current job ccsid (or passed ccsid)
         // to ascii (819) because the checker only understands ascii
         //
         iconv_table = iconv_open(iconv_to : iconv_from);
         if (iconv_table.return_value <> -1);

           length = read(fd : dataPtr : %size(data));
           dow (length >= 1 and valid);
             tmpLength = length;

             if (iconv(iconv_table :
                     dataPtr : tmpLength :
                     json_string_ptr : tmpLength) <> -1);


                              // check string char by char                                                                                        

               for i = 1 to length;                                                                                               

                 char =  %subst(json_string : json_pos + i : 1);                                                                   
                 if (json_checker_char(json_checker_ptr : unsInt) = false);                                                                                          
                   sendDiagnosticMessage('JSON checker aborted at position ' +                                                    
                     %char(i) + ', unicode ' + %char(unsInt) +                                                                    
                     ', character "' + %subst(data : i : 1) + '".');                                                              
                   valid = *off;                                                                                                  
                   leave;                                                                                                          
                 endif;                                                                                                           
                                                                                                                                   
               endfor;                                                                                                            
                                                                                                                                   
               json_pos += length;                                                                                                
                                                                                                                                   
             else;                                                                                                                
               sendDiagnosticMessage('Could not translate buffer from ccsid ' +                                                   
                    %char(iconv_from.ccsid) + ' to ccsid ' +                                                                       
                    %char(iconv_to.ccsid) + '.');                                                                                 
               valid = *off;                                                                                                       
             endif;                                                                                                               
                                                                                                                                   
             dataPtr = %addr(data);                                                                                               
             data = *blanks;                                                                                                       
                                                                                                                                   
             length = read(fd : dataPtr : %size(data));                                                                            
           enddo;                                                                                                                 
                                                                                                                                   
         else;                                                                                                                    
           sendDiagnosticMessage('Could not init iconv.');                                                                         
           valid = *off;                                                                                                          
         endif;                                                                                                                    
                                                                                                                                   
         // free memory                                                                                                            
         if (valid and json_checker_done(json_checker_ptr) = false);                                                              
           valid = *off;                                                                                                           
         endif;                                                                                                                   
                                                                                                                                   
         // cleanup                                                                                                               
         iconv_close(iconv_table);                                                                                                 
         callp close(fd);                                                                                                         
                                                                                                                                   
       else;                                                                                                                      
         sendDiagnosticMessage('Could not open file ' + %trim(filePath) + '.');                                                   
         valid = *off;                                                                                                            
       endif;                                                                                                                     
                                                                                                                                   
       return valid;                                                                                                              
      /end-free                                                                                                                    
     P                 E
  